@@ -4,6 +4,7 @@ import contextlib
 import datetime
 import decimal
 import hashlib
+import inspect
 import itertools
 import json
 import os
@@ -103,6 +104,10 @@ class PrimaryKeyRequired(Exception):
     pass
 
 
+class InvalidColumns(Exception):
+    pass
+
+
 class Database:
     def __init__(
         self,
@@ -143,6 +148,11 @@ class Database:
     def __repr__(self):
         return "<Database {}>".format(self.conn)
 
+    def register_function(self, fn):
+        name = fn.__name__
+        arity = len(inspect.signature(fn).parameters)
+        self.conn.create_function(name, arity, fn)
+
     def execute(self, sql, parameters=None):
         if self._tracer:
             self._tracer(sql, parameters)
@@ -262,7 +272,7 @@ class Database:
                 )
         return fks
 
-    def create_table(
+    def create_table_sql(
         self,
         name,
         columns,
@@ -279,6 +289,8 @@ class Database:
         # any extracts will be treated as integer columns with a foreign key
         extracts = resolve_extracts(extracts)
         for extract_column, extract_table in extracts.items():
+            if isinstance(extract_column, tuple):
+                assert False
             # Ensure other table exists
             if not self[extract_table].exists():
                 self.create_table(extract_table, {"id": int, "value": str}, pk="id")
@@ -320,6 +332,8 @@ class Database:
         column_defs = []
         # ensure pk is a tuple
         single_pk = None
+        if isinstance(pk, list) and len(pk) == 1 and isinstance(pk[0], str):
+            pk = pk[0]
         if isinstance(pk, str):
             single_pk = pk
             if pk not in [c[0] for c in column_items]:
@@ -330,7 +344,7 @@ class Database:
                 column_extras.append("PRIMARY KEY")
             if column_name in not_null:
                 column_extras.append("NOT NULL")
-            if column_name in defaults:
+            if column_name in defaults and defaults[column_name] is not None:
                 column_extras.append(
                     "DEFAULT {}".format(self.escape(defaults[column_name]))
                 )
@@ -362,6 +376,31 @@ class Database:
         """.format(
             table=name, columns_sql=columns_sql, extra_pk=extra_pk
         )
+        return sql
+
+    def create_table(
+        self,
+        name,
+        columns,
+        pk=None,
+        foreign_keys=None,
+        column_order=None,
+        not_null=None,
+        defaults=None,
+        hash_id=None,
+        extracts=None,
+    ):
+        sql = self.create_table_sql(
+            name=name,
+            columns=columns,
+            pk=pk,
+            foreign_keys=foreign_keys,
+            column_order=column_order,
+            not_null=not_null,
+            defaults=defaults,
+            hash_id=hash_id,
+            extracts=extracts,
+        )
         self.execute(sql)
         return self.table(
             name,
@@ -497,10 +536,10 @@ class Queryable:
     def rows(self):
         return self.rows_where()
 
-    def rows_where(self, where=None, where_args=None, order_by=None):
+    def rows_where(self, where=None, where_args=None, order_by=None, select="*"):
         if not self.exists():
             return []
-        sql = "select * from [{}]".format(self.name)
+        sql = "select {} from [{}]".format(select, self.name)
         if where is not None:
             sql += " where " + where
         if order_by is not None:
@@ -685,6 +724,256 @@ class Table(Queryable):
             )
         return self
 
+    def transform(
+        self,
+        *,
+        types=None,
+        rename=None,
+        drop=None,
+        pk=DEFAULT,
+        not_null=None,
+        defaults=None,
+        drop_foreign_keys=None,
+        column_order=None,
+    ):
+        assert self.exists(), "Cannot transform a table that doesn't exist yet"
+        sqls = self.transform_sql(
+            types=types,
+            rename=rename,
+            drop=drop,
+            pk=pk,
+            not_null=not_null,
+            defaults=defaults,
+            drop_foreign_keys=drop_foreign_keys,
+            column_order=column_order,
+        )
+        pragma_foreign_keys_was_on = self.db.execute("PRAGMA foreign_keys").fetchone()[
+            0
+        ]
+        try:
+            if pragma_foreign_keys_was_on:
+                self.db.execute("PRAGMA foreign_keys=0;")
+            with self.db.conn:
+                for sql in sqls:
+                    self.db.execute(sql)
+                # Run the foreign_key_check before we commit
+                if pragma_foreign_keys_was_on:
+                    self.db.execute("PRAGMA foreign_key_check;")
+        finally:
+            if pragma_foreign_keys_was_on:
+                self.db.execute("PRAGMA foreign_keys=1;")
+        return self
+
+    def transform_sql(
+        self,
+        *,
+        types=None,
+        rename=None,
+        drop=None,
+        pk=DEFAULT,
+        not_null=None,
+        defaults=None,
+        drop_foreign_keys=None,
+        column_order=None,
+        tmp_suffix=None,
+    ):
+        types = types or {}
+        rename = rename or {}
+        drop = drop or set()
+        new_table_name = "{}_new_{}".format(
+            self.name, tmp_suffix or os.urandom(6).hex()
+        )
+        current_column_pairs = list(self.columns_dict.items())
+        new_column_pairs = []
+        copy_from_to = {column: column for column, _ in current_column_pairs}
+        for name, type_ in current_column_pairs:
+            type_ = types.get(name) or type_
+            if name in drop:
+                del [copy_from_to[name]]
+                continue
+            new_name = rename.get(name) or name
+            new_column_pairs.append((new_name, type_))
+            copy_from_to[name] = new_name
+
+        sqls = []
+        if pk is DEFAULT:
+            pks_renamed = tuple(rename.get(p) or p for p in self.pks)
+            if len(pks_renamed) == 1:
+                pk = pks_renamed[0]
+            else:
+                pk = pks_renamed
+
+        # not_null may be a set or dict, need to convert to a set
+        create_table_not_null = {
+            rename.get(c.name) or c.name
+            for c in self.columns
+            if c.notnull
+            if c.name not in drop
+        }
+        if isinstance(not_null, dict):
+            # Remove any columns with a value of False
+            for key, value in not_null.items():
+                # Column may have been renamed
+                key = rename.get(key) or key
+                if value is False and key in create_table_not_null:
+                    create_table_not_null.remove(key)
+                else:
+                    create_table_not_null.add(key)
+        elif isinstance(not_null, set):
+            create_table_not_null.update((rename.get(k) or k) for k in not_null)
+        elif not_null is None:
+            pass
+        else:
+            assert False, "not_null must be a dict or a set or None"
+        # defaults=
+        create_table_defaults = {
+            (rename.get(c.name) or c.name): c.default_value
+            for c in self.columns
+            if c.default_value is not None and c.name not in drop
+        }
+        if defaults is not None:
+            create_table_defaults.update(
+                {rename.get(c) or c: v for c, v in defaults.items()}
+            )
+
+        # foreign_keys
+        create_table_foreign_keys = []
+        for table, column, other_table, other_column in self.foreign_keys:
+            if (drop_foreign_keys is None) or (column not in drop_foreign_keys):
+                create_table_foreign_keys.append(
+                    (rename.get(column) or column, other_table, other_column)
+                )
+
+        if column_order is not None:
+            column_order = [rename.get(col) or col for col in column_order]
+
+        sqls.append(
+            self.db.create_table_sql(
+                new_table_name,
+                dict(new_column_pairs),
+                pk=pk,
+                not_null=create_table_not_null,
+                defaults=create_table_defaults,
+                foreign_keys=create_table_foreign_keys,
+                column_order=column_order,
+            ).strip()
+        )
+
+        # Copy across data, respecting any renamed columns
+        new_cols = []
+        old_cols = []
+        for from_, to_ in copy_from_to.items():
+            old_cols.append(from_)
+            new_cols.append(to_)
+        copy_sql = "INSERT INTO [{new_table}] ({new_cols})\n   SELECT {old_cols} FROM [{old_table}];".format(
+            new_table=new_table_name,
+            old_table=self.name,
+            old_cols=", ".join("[{}]".format(col) for col in old_cols),
+            new_cols=", ".join("[{}]".format(col) for col in new_cols),
+        )
+        sqls.append(copy_sql)
+        # Drop the old table
+        sqls.append("DROP TABLE [{}];".format(self.name))
+        # Rename the new one
+        sqls.append(
+            "ALTER TABLE [{}] RENAME TO [{}];".format(new_table_name, self.name)
+        )
+        return sqls
+
+    def extract(self, columns, table=None, fk_column=None, rename=None):
+        rename = rename or {}
+        if isinstance(columns, str):
+            columns = [columns]
+        if not set(columns).issubset(self.columns_dict.keys()):
+            raise InvalidColumns(
+                "Invalid columns {} for table with columns {}".format(
+                    columns, list(self.columns_dict.keys())
+                )
+            )
+        table = table or "_".join(columns)
+        first_column = columns[0]
+        pks = self.pks
+        lookup_table = self.db[table]
+        fk_column = fk_column or "{}_id".format(table)
+        magic_lookup_column = "{}_{}".format(fk_column, os.urandom(6).hex())
+
+        # Populate the lookup table with all of the extracted unique values
+        lookup_columns_definition = {
+            (rename.get(col) or col): typ
+            for col, typ in self.columns_dict.items()
+            if col in columns
+        }
+        if lookup_table.exists():
+            if not set(lookup_columns_definition.items()).issubset(
+                lookup_table.columns_dict.items()
+            ):
+                raise InvalidColumns(
+                    "Lookup table {} already exists but does not have columns {}".format(
+                        table, lookup_columns_definition
+                    )
+                )
+        else:
+            lookup_table.create(
+                {
+                    **{
+                        "id": int,
+                    },
+                    **lookup_columns_definition,
+                },
+                pk="id",
+            )
+        lookup_columns = [(rename.get(col) or col) for col in columns]
+        lookup_table.create_index(lookup_columns, unique=True, if_not_exists=True)
+        self.db.execute(
+            "INSERT OR IGNORE INTO [{lookup_table}] ({lookup_columns}) SELECT DISTINCT {table_cols} FROM [{table}]".format(
+                lookup_table=table,
+                lookup_columns=", ".join("[{}]".format(c) for c in lookup_columns),
+                table_cols=", ".join("[{}]".format(c) for c in columns),
+                table=self.name,
+            )
+        )
+
+        # Now add the new fk_column
+        self.add_column(magic_lookup_column, int)
+
+        # And populate it
+        self.db.execute(
+            "UPDATE [{table}] SET [{magic_lookup_column}] = (SELECT id FROM [{lookup_table}] WHERE {where})".format(
+                table=self.name,
+                magic_lookup_column=magic_lookup_column,
+                lookup_table=table,
+                where=" AND ".join(
+                    "[{table}].[{column}] = [{lookup_table}].[{lookup_column}]".format(
+                        table=self.name,
+                        lookup_table=table,
+                        column=column,
+                        lookup_column=rename.get(column) or column,
+                    )
+                    for column in columns
+                ),
+            )
+        )
+        # Figure out the right column order
+        column_order = []
+        for c in self.columns:
+            if c.name in columns and magic_lookup_column not in column_order:
+                column_order.append(magic_lookup_column)
+            elif c.name == magic_lookup_column:
+                continue
+            else:
+                column_order.append(c.name)
+
+        # Drop the unnecessary columns and rename lookup column
+        self.transform(
+            drop=set(columns),
+            rename={magic_lookup_column: fk_column},
+            column_order=column_order,
+        )
+
+        # And add the foreign key constraint
+        self.add_foreign_key(fk_column, table, "id")
+        return self
+
     def create_index(self, columns, index_name=None, unique=False, if_not_exists=False):
         if index_name is None:
             index_name = "idx_{}_{}".format(
@@ -781,7 +1070,9 @@ class Table(Queryable):
         else:
             return pks[0].name
 
-    def add_foreign_key(self, column, other_table=None, other_column=None):
+    def add_foreign_key(
+        self, column, other_table=None, other_column=None, ignore=False
+    ):
         # Ensure column exists
         if column not in self.columns_dict:
             raise AlterError("No such column: {}".format(column))
@@ -806,24 +1097,33 @@ class Table(Queryable):
             and fk.other_table == other_table
             and fk.other_column == other_column
         ):
-            raise AlterError(
-                "Foreign key already exists for {} => {}.{}".format(
-                    column, other_table, other_column
+            if ignore:
+                return self
+            else:
+                raise AlterError(
+                    "Foreign key already exists for {} => {}.{}".format(
+                        column, other_table, other_column
+                    )
                 )
-            )
         self.db.add_foreign_keys([(self.name, column, other_table, other_column)])
+        return self
 
     def enable_fts(
-        self, columns, fts_version="FTS5", create_triggers=False, tokenize=None
+        self,
+        columns,
+        fts_version="FTS5",
+        create_triggers=False,
+        tokenize=None,
+        replace=False,
     ):
         "Enables FTS on the specified columns."
-        sql = (
+        create_fts_sql = (
             textwrap.dedent(
                 """
             CREATE VIRTUAL TABLE [{table}_fts] USING {fts_version} (
                 {columns},{tokenize}
                 content=[{table}]
-            );
+            )
         """
             )
             .strip()
@@ -834,7 +1134,25 @@ class Table(Queryable):
                 tokenize="\n    tokenize='{}',".format(tokenize) if tokenize else "",
             )
         )
-        self.db.executescript(sql)
+        should_recreate = False
+        if replace and self.db["{}_fts".format(self.name)].exists():
+            # Does the table need to be recreated?
+            fts_schema = self.db["{}_fts".format(self.name)].schema
+            if fts_schema != create_fts_sql:
+                should_recreate = True
+            expected_triggers = {self.name + suffix for suffix in ("_ai", "_ad", "_au")}
+            existing_triggers = {t.name for t in self.triggers}
+            has_triggers = existing_triggers.issuperset(expected_triggers)
+            if has_triggers != create_triggers:
+                should_recreate = True
+            if not should_recreate:
+                # Table with correct configuration already exists
+                return self
+
+        if should_recreate:
+            self.disable_fts()
+
+        self.db.executescript(create_fts_sql)
         self.populate_fts(columns)
 
         if create_triggers:
@@ -904,6 +1222,7 @@ class Table(Queryable):
         with self.db.conn:
             for trigger_name in trigger_names:
                 self.db.execute("DROP TRIGGER IF EXISTS [{}]".format(trigger_name))
+        return self
 
     def rebuild_fts(self):
         fts_table = self.detect_fts()
@@ -915,6 +1234,7 @@ class Table(Queryable):
                 table=fts_table
             )
         )
+        return self
 
     def detect_fts(self):
         "Detect if table has a corresponding FTS virtual table and return it"
@@ -982,6 +1302,7 @@ class Table(Queryable):
         )
         with self.db.conn:
             self.db.execute(sql, pk_values)
+        return self
 
     def delete_where(self, where=None, where_args=None):
         if not self.exists():
@@ -990,6 +1311,7 @@ class Table(Queryable):
         if where is not None:
             sql += " where " + where
         self.db.execute(sql, where_args or [])
+        return self
 
     def update(self, pk_values, updates=None, alter=False, conversions=None):
         updates = updates or {}
@@ -1003,11 +1325,12 @@ class Table(Queryable):
         args = []
         sets = []
         wheres = []
+        pks = self.pks
         validate_column_names(updates.keys())
         for key, value in updates.items():
             sets.append("[{}] = {}".format(key, conversions.get(key, "?")))
             args.append(value)
-        wheres = ["[{}] = ?".format(pk_name) for pk_name in self.pks]
+        wheres = ["[{}] = ?".format(pk_name) for pk_name in pks]
         args.extend(pk_values)
         sql = "update [{table}] set {sets} where {wheres}".format(
             table=self.name, sets=", ".join(sets), wheres=" and ".join(wheres)
@@ -1025,7 +1348,7 @@ class Table(Queryable):
 
             # TODO: Test this works (rolls back) - use better exception:
             assert rowcount == 1
-        self.last_pk = pk_values[0] if len(self.pks) == 1 else pk_values
+        self.last_pk = pk_values[0] if len(pks) == 1 else pk_values
         return self
 
     def build_insert_queries_and_params(
@@ -1423,6 +1746,7 @@ class Table(Queryable):
         for col_name, col_type in needed_columns.items():
             if col_name not in current_columns:
                 self.add_column(col_name, col_type)
+        return self
 
     def lookup(self, column_values):
         # lookups is a dictionary - all columns will be used for a unique index
